#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <libxml/tree.h>

#include "ruleset.h"
#include "mainwin.h"
#include "support.h"
#include "format.h"
#include "interface.h"

gchar* ruleset_file = NULL;
gboolean ruleset_modified = FALSE;
static GList *ruleset = NULL;

static pid_t child_pid;

static GtkListStore *ruleset_list_store = NULL;
enum { COLUMN_ENABLED, COLUMN_DESCRIPTION, COLUMN_VERDICT, COLUMN_RULE, N_COLUMNS };

static int commit_pipe[2] = {-1, -1};
volatile static gboolean commit_running = FALSE;
GtkWidget *commit_window = NULL;

void ruleset_free() {
    while (ruleset) {
        rule_free(ruleset->data);
        ruleset = g_list_remove(ruleset, ruleset->data);
    }

    ruleset_modified = FALSE;
    g_free(ruleset_file);
    ruleset_file = NULL;
}


void ruleset_clear() {
    ruleset_free();
    gtk_list_store_clear(ruleset_list_store);
    ruleset_update_ui();
}


void _set_data(rule_t *rule) {

    gtk_list_store_set(ruleset_list_store, &rule->iter,
                       COLUMN_ENABLED, rule->enabled,
                       COLUMN_DESCRIPTION, rule->description,
                       COLUMN_VERDICT, format_verdict(rule->verdict, FORMAT_USER),
                       COLUMN_RULE, rule,
                       -1);
}

void ruleset_fill_ui() {
    GList *l;

    gtk_list_store_clear(ruleset_list_store);

    for (l = ruleset; l; l = l->next) {
        rule_t *rule = (rule_t*) l->data;
        gtk_list_store_append(ruleset_list_store, &rule->iter);
        _set_data(rule);
        rule->realized = TRUE;
    }
    
    ruleset_update_ui();
}

int ruleset_load(gchar *fn) {
    xmlDocPtr doc = NULL;
    xmlNodePtr node;
    
    g_assert(!ruleset && !ruleset_file);

    if (!(doc = xmlParseFile(fn)))
        goto finish;

    if (!(node = xmlDocGetRootElement(doc)))
        goto finish;

    if (xmlStrcmp(node->name, "ruleset"))
        goto finish;

    for (node = node->xmlChildrenNode; node; node = node->next) {
        rule_t *rule;

        if (node->type == XML_TEXT_NODE)
            continue;

        if (node->type != XML_ELEMENT_NODE)
            goto finish;
        
        if (xmlStrcmp(node->name, "rule"))
            goto finish;

        if (!(rule = rule_new_from_xml(doc, node)))
            goto finish;

        ruleset = g_list_append(ruleset, rule);
    }

    xmlFreeDoc(doc);

    ruleset_file = g_strdup(fn);
    ruleset_modified = FALSE;
    ruleset_fill_ui();
    
    return 0;

finish:
    fprintf(stderr, "Broken XML file.\n");

    if (doc)
        xmlFreeDoc(doc);

    ruleset_free();
    
    return -1;
}

int ruleset_save(gchar *fn) {
    GList *l;
    xmlDocPtr doc = NULL;
    xmlNodePtr node;
    int r = -1;

    if (!(doc = xmlNewDoc("1.0")))
        goto finish;

    if (!(node = xmlNewNode(NULL, "ruleset")))
        goto finish;
    
    xmlDocSetRootElement(doc, node);
    
    for (l = ruleset; l; l = l->next) {
        if (rule_to_xml((rule_t*) l->data, doc, node) < 0)
            goto finish;
    }

    if (xmlSaveFormatFile(fn, doc, 1) < 0)
        goto finish;

    xmlFreeDoc(doc);

    ruleset_modified = FALSE;
    ruleset_fill_ui();
        
    return 0;

finish:
    fprintf(stderr, "Could not write XML file.\n");
    
    if (doc)
        xmlFreeDoc(doc);
    
    return r;
}

rule_t* ruleset_get_current_rule() {
    GtkTreePath *path;
    GtkTreeIter iter;
    GtkTreeView *tv = GTK_TREE_VIEW(lookup_widget(get_main_window(), "ruleset_view"));
    rule_t *rule;
    
    gtk_tree_view_get_cursor(tv, &path, NULL);

    if (!path)
        return NULL;
    
    gtk_tree_model_get_iter(GTK_TREE_MODEL(ruleset_list_store), &iter, path);
    gtk_tree_model_get(GTK_TREE_MODEL(ruleset_list_store), &iter, COLUMN_RULE, &rule, -1);
    
    return rule;
    
}

int ruleset_install() {
    char *argv[3];
    GError *e;

    g_assert(ruleset_file);
    
    argv[0] = "../helper/install-firewall";
    argv[1] = ruleset_file;
    argv[2] = NULL;

    if (!g_spawn_async_with_pipes(NULL, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &child_pid, NULL, NULL, NULL, &e)) {
        fprintf(stderr, "Could not run install-firewall: %s\n", e->message);
        g_error_free(e);
        return -1;
    }

    return 0;
}

static void _sigchld(int sig) {
    gchar c = 'X';
    write(commit_pipe[1], &c, 1);
    signal(SIGCHLD, SIG_DFL);
}


static gboolean _commit_finish(GIOChannel *source, GIOCondition condition, gpointer data) {
    gchar c;
    read(commit_pipe[0], &c, 1);
    gtk_widget_hide(commit_window);
    commit_running = FALSE;
    return TRUE;
}

int ruleset_commit() {
    g_assert(ruleset_file);

    if (commit_running)
        return 0;
    
    commit_running = TRUE;

    if (!commit_window)
        commit_window = create_wait_window();
    
    gtk_widget_show_all(commit_window);

    signal(SIGCHLD, _sigchld);
    
    if (ruleset_modified)
        ruleset_save(ruleset_file);
    
    return ruleset_install();
}

void ruleset_update_ui() {
    GList *l;
    rule_t *rule = ruleset_get_current_rule();
    gtk_widget_set_sensitive(GTK_WIDGET(lookup_widget(get_main_window(), "properties_button")), rule ? TRUE : FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(lookup_widget(get_main_window(), "remove_button")), rule ? TRUE : FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(lookup_widget(get_main_window(), "clear_button")), ruleset ? TRUE : FALSE);

    l = g_list_first(ruleset);
    gtk_widget_set_sensitive(GTK_WIDGET(lookup_widget(get_main_window(), "up_button")), !rule || !l || l->data == rule ? FALSE : TRUE);
    l = g_list_last(ruleset);
    gtk_widget_set_sensitive(GTK_WIDGET(lookup_widget(get_main_window(), "down_button")), !rule || !l || l->data == rule ? FALSE : TRUE);


    gtk_widget_set_sensitive(GTK_WIDGET(lookup_widget(get_main_window(), "commit_button")), ruleset_modified);
    
    if (rule) {
        gtk_label_set_label(GTK_LABEL(lookup_widget(get_main_window(), "rule_name_label")), rule->description);
        gtk_label_set_label(GTK_LABEL(lookup_widget(get_main_window(), "rule_match_label")), rule_match_string(rule));
        gtk_label_set_label(GTK_LABEL(lookup_widget(get_main_window(), "rule_verdict_label")), format_verdict(rule->verdict, FORMAT_USER));
    } else {
        gchar *p = "[n/a]";
        gtk_label_set_label(GTK_LABEL(lookup_widget(get_main_window(), "rule_name_label")), p);
        gtk_label_set_label(GTK_LABEL(lookup_widget(get_main_window(), "rule_match_label")), p);
        gtk_label_set_label(GTK_LABEL(lookup_widget(get_main_window(), "rule_verdict_label")), p);
    }

}

int ruleset_append_rule(rule_t *rule) {
    g_assert(rule);
    
    if (g_list_find(ruleset, rule))
        return -1;
    
    ruleset = g_list_append(ruleset, rule);
    gtk_list_store_append(ruleset_list_store, &rule->iter);
    _set_data(rule);

    rule->realized = TRUE;

    ruleset_modified = TRUE;
    ruleset_update_ui();
    return 0;
}

int ruleset_update_rule(rule_t *rule) {
    g_assert(rule && rule->realized);

    _set_data(rule);
    
    ruleset_modified = TRUE;
    ruleset_update_ui();
    return 0;
}


int ruleset_remove_rule(rule_t *rule) {
    g_assert(rule);
    
    if (!g_list_find(ruleset, rule))
        return -1;

    gtk_list_store_remove(ruleset_list_store, &rule->iter);
    ruleset = g_list_remove(ruleset, rule);
    rule_free(rule);
    
    ruleset_modified = TRUE;
    ruleset_update_ui();
    return 0;
}


int ruleset_move_rule(rule_t *rule, int i) {
    GtkTreeIter *iter = NULL;
    GList *a = NULL, *l;
    
    if (i == 0)
        return -1;

    g_message("foo %i", i);
    ruleset_modified = TRUE;

    l = g_list_find(ruleset, rule);
    g_assert(l);

    
    if (i > 0) {
        if (!l->next)
            return -1;

        iter = &(((rule_t*) l->next)->iter);

        a = l->next->next;
    }

    if (i < 0) {
        if (!l->prev)
            return -1;

        iter = &(((rule_t*) l->prev)->iter);
        
        a = l->prev;
    }

    ruleset = g_list_remove(ruleset, rule);
    ruleset = g_list_insert_before(ruleset, a, rule);

    g_error("Waiting for GTK 2.2.1");
    
/*     if (i < 0) */
/*         gtk_list_store_insert_before(ruleset_list_store, &rule->iter, iter); */
/*     else */
/*         gtk_list_store_insert_after(ruleset_list_store, &rule->iter, iter); */
        
    return 0;
}

void ruleset_widget_init() {
    GtkTreeView *tv = GTK_TREE_VIEW(lookup_widget(get_main_window(), "ruleset_view"));
    GtkTreeViewColumn *c;

    ruleset_list_store = gtk_list_store_new(N_COLUMNS, G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);
    gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ruleset_list_store));

    c = gtk_tree_view_column_new_with_attributes("", gtk_cell_renderer_toggle_new(), "active", COLUMN_ENABLED, NULL);
    gtk_tree_view_append_column(tv, c);

    c = gtk_tree_view_column_new_with_attributes("Description", gtk_cell_renderer_text_new(), "text", COLUMN_DESCRIPTION, NULL);
    gtk_tree_view_column_set_resizable(c, TRUE);
    gtk_tree_view_append_column(tv, c);

    c = gtk_tree_view_column_new_with_attributes("Verdict", gtk_cell_renderer_text_new(), "text", COLUMN_VERDICT, NULL);
    gtk_tree_view_column_set_resizable(c, TRUE);
    gtk_tree_view_append_column(tv, c);

    ruleset_update_ui();
}


int ruleset_init() {
    GIOChannel *c;
    if (pipe(commit_pipe) < 0)
        return -1;

    if (!(c = g_io_channel_unix_new(commit_pipe[0]))) {
        ruleset_done();
        return -1;
    }
    
    g_io_add_watch(c, G_IO_IN, _commit_finish, NULL);
    g_io_channel_unref(c);

    return 0;
}

void ruleset_done() {
    ruleset_free();
    if (commit_pipe[0] >= 0)
        close(commit_pipe[0]);
    if (commit_pipe[1] >= 0)
        close(commit_pipe[1]);
    commit_pipe[0] = commit_pipe[1] = -1;
}
